from typing import List, Tuple

from rdkit import Chem
from rdkit.Chem import AllChem, MolFromSmiles, MolToSmiles
from rdkit.Chem.rdchem import Mol
from rdkit.DataStructs.cDataStructs import UIntSparseIntVect


class GeneralChemistry:

    @staticmethod
    def smiles_to_mols_and_indices(query_smiles: List[str]) -> Tuple[List[Mol], List[int]]:
        mols = [Chem.MolFromSmiles(smile) for smile in query_smiles]
        valid_mask = [mol is not None for mol in mols]
        valid_idxs = [idx for idx, is_valid in enumerate(valid_mask) if is_valid]
        valid_mols = [mols[idx] for idx in valid_idxs]
        return valid_mols, valid_idxs

    @staticmethod
    def mols_to_fingerprints(molecules: List[Mol], radius=3, use_counts=True, use_features=True) \
            -> List[UIntSparseIntVect]:
        fingerprints = [AllChem.GetMorganFingerprint(mol, radius, useCounts=use_counts, useFeatures=use_features) for
                        mol in molecules]
        return fingerprints

    @staticmethod
    def smiles_to_mols(query_smiles: List[str]) -> List[Mol]:
        mols = [Chem.MolFromSmiles(smile) for smile in query_smiles]
        valid_mask = [mol is not None for mol in mols]
        valid_idxs = [idx for idx, is_valid in enumerate(valid_mask) if is_valid]
        valid_mols = [mols[idx] for idx in valid_idxs]
        return valid_mols

    def smiles_to_fingerprints(self, query_smiles: List[str], radius=3, use_counts=True, use_features=True) -> List[
        UIntSparseIntVect]:
        mols = self.smiles_to_mols(query_smiles)
        fingerprints = self.mols_to_fingerprints(mols, radius=radius, use_counts=use_counts, use_features=use_features)
        return fingerprints

    def smile_to_mol(self, smile: str) -> Mol:
        """
        Creates a Mol object from a SMILES string.
        :param smile: SMILES string.
        :return: A Mol object or None if it's not valid.
        """
        if smile:
            return MolFromSmiles(smile)

    def mols_to_smiles(self, molecules: List[Mol]) -> List[str]:
        """This method assumes that all molecules are valid."""
        valid_smiles = [MolToSmiles(mol, isomericSmiles=False) for mol in molecules]
        return valid_smiles



import json
import os
from abc import abstractmethod
from typing import List

import numpy as np

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.enums import EnvironmentalVariablesEnum
from reinvent_scoring.scoring.score_components.pip.base_rest_component import BaseRESTComponent


class BasePiPModelComponent(BaseRESTComponent):
    def __init__(self, parameters: ComponentParameters):
        self._environment_keys = EnvironmentalVariablesEnum()
        super().__init__(parameters)

    def _parse_response(self, response_json: dict, data_size: int) -> np.array:
        compounds = response_json['jsonData']['data']
        results_raw = np.empty(data_size, dtype=np.float32)
        results_raw[:] = np.nan

        try:
            for compound in compounds:
                try:
                    index = int(compound["id"])
                    results_raw[index] = self._parse_single_compound(compound)

                except (ValueError, TypeError, KeyError):
                    pass  # If parsing failed, keep value NaN for this compound and continue.
        finally:
            return results_raw

    @abstractmethod
    def _parse_single_compound(self, compound):
        raise NotImplementedError("_parse_compound method is not implemented")

    def _format_data(self, smiles: List[str]) -> dict:
        molecules = [{"molData": smi, "id": f"{i}"} for i, smi in enumerate(smiles)]
        data = {
            "jsonData": {
                "data": molecules,
                "metadata": {
                    "molFormat":
                        "smiles"
                },
                "parameters": {}
            }
        }
        return data

    def _create_url(self, component_name) -> str:
        pip_url = self._get_enviornment_variable(self._environment_keys.PIP_URL)
        request_url = pip_url.format(component_name)
        return request_url

    def _create_header(self) -> dict:
        pip_key = self._get_enviornment_variable(self._environment_keys.PIP_KEY)

        header = {'Content-Type': 'application/json', 'x-api-key': pip_key}
        return header

    def _get_enviornment_variable(self, variable: str) -> str:
        try:
            return os.environ[variable]
        except KeyError:
            return self._retrieve_pip_key_from_config(variable)

    def _retrieve_pip_key_from_config(self, variable: str) -> str:
        try:
            project_root = os.path.dirname(__file__)
            with open(os.path.join(project_root, '../../../configs/config.json'), 'r') as f:
                config = json.load(f)
            environmental_variables = config[self._environment_keys.ENVIRONMENTAL_VARIABLES]
            return environmental_variables[variable]
        except KeyError as ex:
            raise KeyError(f"Key {variable} not found in reinvent_scoring/configs/config.json")


from typing import List, Dict

import numpy as np
from rdkit import DataStructs
from rdkit.Avalon import pyAvalonTools
from rdkit.Chem import AllChem, MACCSkeys, Mol
from rdkit.Chem.rdMolDescriptors import GetHashedMorganFingerprint

from reinvent_scoring.scoring.enums import ComponentSpecificParametersEnum
from reinvent_scoring.scoring.enums import DescriptorTypesEnum


class Descriptors:
    """Molecular descriptors.

    The descriptors in this class are mostly RDKit fingerprints used as inputs to
    scikit-learn predictive models. Since scikit-learn predictive models take
    numpy arrays as input, RDKit fingerprints are converted to numpy arrays.
    """

    def __init__(self):
        self._descriptor_types = DescriptorTypesEnum()
        self._specific_parameters = ComponentSpecificParametersEnum()

    def load_descriptor(self, parameters: {}):
        descriptor_type = parameters.get(self._specific_parameters.DESCRIPTOR_TYPE, self._descriptor_types.ECFP_COUNTS)
        registry = self._descriptor_registry()
        descriptor = registry[descriptor_type]
        return descriptor

    def _descriptor_registry(self) -> dict:
        descriptor_list = dict(ecfp=self.molecules_to_fingerprints,
                               ecfp_counts=self.molecules_to_count_fingerprints,
                               maccs_keys=self.maccs_keys,
                               avalon=self.avalon)
        return descriptor_list

    def maccs_keys(self, molecules: List[Mol], parameters: {}):
        fps = [MACCSkeys.GenMACCSKeys(mol) for mol in molecules]
        fingerprints = [self._numpy_fingerprint(fp, dtype=np.int32) for fp in fps]
        return fingerprints

    def avalon(self, molecules: List[Mol], parameters: {}):
        size = parameters.get('size', 512)
        fps = [pyAvalonTools.GetAvalonFP(mol, nBits=size) for mol in molecules]
        fingerprints = [self._numpy_fingerprint(fp, dtype=np.int32) for fp in fps]
        return fingerprints

    def molecules_to_fingerprints(self, molecules: List[Mol], parameters: {}) -> List[np.ndarray]:
        radius = parameters.get('radius', 3)
        size = parameters.get('size', 2048)
        fp_bits = [AllChem.GetMorganFingerprintAsBitVect(mol, radius, size) for mol in molecules]
        fingerprints = [self._numpy_fingerprint(fp, dtype=np.int32) for fp in fp_bits]
        return fingerprints

    def molecules_to_count_fingerprints_ori(self, molecules: List[Mol], parameters: {}) -> np.ndarray:
        """Morgan Count Fingerprints.

        This is "original" implementation, with own hashing code.
        """
        radius = parameters.get('radius', 3)
        useCounts = parameters.get('use_counts', True)
        useFeatures = parameters.get('use_features', True)
        size = parameters.get('size', 2048)
        fps = [AllChem.GetMorganFingerprint(mol, radius, useCounts=useCounts, useFeatures=useFeatures) for mol in molecules]
        nfp = np.zeros((len(fps), size), np.int32)
        for i, fp in enumerate(fps):
            for idx, v in fp.GetNonzeroElements().items():
                nidx = idx % size
                nfp[i, nidx] += int(v)
        return nfp

    def molecules_to_count_fingerprints(self, molecules: List[Mol], parameters: Dict) -> List[np.ndarray]:
        """Morgan Count Fingerprints from RDKit.

        This implementation uses RDKit's hashing through GetHashedMorganFingerprint.
        See https://stackoverflow.com/a/55119975
        """

        radius = parameters.get("radius", 3)
        useFeatures = parameters.get("use_features", True)
        size = parameters.get("size", 2048)
        fps = [
            GetHashedMorganFingerprint(
                mol, radius, nBits=size, useFeatures=useFeatures,
            )
            for mol in molecules
        ]
        fingerprints = [self._numpy_fingerprint(fp, dtype=np.int32) for fp in fps]
        return fingerprints

    def _numpy_fingerprint(self, rdkit_fingerprint, dtype=None) -> np.ndarray:
        """Converts RDKit fingerprint to numpy array."""

        numpy_fp = np.zeros((0,), dtype=dtype)  # Initialize empty array, RDKit will resize it.
        DataStructs.ConvertToNumpyArray(rdkit_fingerprint, numpy_fp)
        return numpy_fp



import asyncio
from abc import abstractmethod
from typing import List

import aiohttp
# import nest_asyncio
import numpy as np
import rdkit.Chem as rkc

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.enums import TransformationTypeEnum
from reinvent_scoring.scoring.score_components import BaseScoreComponent
from reinvent_scoring.scoring.score_summary import ComponentSummary
from reinvent_scoring.scoring.score_transformations import TransformationFactory


# nest_asyncio.apply()


class BaseRESTComponent(BaseScoreComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)
        # self._environment_keys = EnvironmentalVariablesEnum()
        self.transformation_function = self._assign_transformation(self.parameters.specific_parameters)
        self._request_url = self._create_url(self.parameters.component_type)
        self._request_header = self._create_header()

    def calculate_score(self, molecules: List[rkc.Mol]) -> ComponentSummary:
        valid_smiles = self._chemistry.mols_to_smiles(molecules)
        score, raw_score = self._score_smiles(valid_smiles)
        score_summary = ComponentSummary(total_score=score, parameters=self.parameters, raw_score=raw_score)

        return score_summary

    def _score_smiles(self, smiles: List[str]) -> np.array:
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(self._post_request(self._request_url, smiles, self._request_header))
        results_raw = self._parse_response(response, len(smiles))
        results = self._apply_score_transformation(results_raw)

        return results, results_raw

    async def _post_request(self, url, smiles, header):
        data = self._format_data(smiles)
        result = await self._execute_request(url, data, header)

        return result

    async def _execute_request(self, request_url, data, header) -> str:
        async with aiohttp.ClientSession() as session:
            async with session.post(request_url, json=data, headers=header) as response:
                if response.status != 200:
                    raise ValueError(
                        f" Status: {response.status} Reason: ({response.reason})."
                        f"Response content: {response.content}"
                    )
                return await response.json()

    @abstractmethod
    def _parse_response(self, response_json: dict, data_size: int) -> np.array:
        raise NotImplementedError("_parse_response method is not implemented")

    def _apply_score_transformation(self, results_raw: np.array) -> np.array:
        """Returns np.array with non-NaN elements transformed by transformation function, and all NaN elements
        transformed into 0. """
        valid_mask = ~np.isnan(results_raw)
        results_raw_valid = results_raw[valid_mask]
        results_transformed = self.transformation_function(results_raw_valid, self.parameters.specific_parameters)
        results = np.zeros(len(results_raw), dtype=np.float32)
        results[valid_mask] = results_transformed

        return results

    @abstractmethod
    def _parse_single_compound(self, compound):
        raise NotImplementedError("_parse_compound method is not implemented")

    @abstractmethod
    def _format_data(self, smiles: List[str]) -> dict:
        raise NotImplementedError("_format_data method is not implemented")

    @abstractmethod
    def _create_url(self, component_name) -> str:
        raise NotImplementedError("_create_url method is not implemented")

    @abstractmethod
    def _create_header(self) -> dict:
        raise NotImplementedError("_create_header method is not implemented")

    def _assign_transformation(self, specific_parameters: {}):
        transformation_type = TransformationTypeEnum()
        factory = TransformationFactory()
        if self.parameters.specific_parameters[self.component_specific_parameters.TRANSFORMATION]:
            transform_function = factory.get_transformation_function(specific_parameters)
        else:
            self.parameters.specific_parameters[
                self.component_specific_parameters.TRANSFORMATION_TYPE] = transformation_type.NO_TRANSFORMATION
            transform_function = factory.no_transformation

        return transform_function
